@@ -1,34 +1,24 @@
-# coding: utf-8
 
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
 
 """Base Widget class.  Allows user to create widgets in the back-end that render
-in the IPython notebook front-end.
+in the Jupyter notebook front-end.
 """
-
 import os
 from contextlib import contextmanager
-try:
-    from collections.abc import Iterable
-except ImportError:
-    # Python 2.7
-    from collections import Iterable
-
-from IPython.core.getipython import get_ipython
+from collections.abc import Iterable
+from IPython import get_ipython
 from ipykernel.comm import Comm
 from traitlets import (
     HasTraits, Unicode, Dict, Instance, List, Int, Set, Bytes, observe, default, Container,
     Undefined)
-from ipython_genutils.py3compat import string_types, PY3
-from IPython.display import display
 from json import loads as jsonloads, dumps as jsondumps
 
 from base64 import standard_b64encode
 
 from .._version import __protocol_version__, __control_protocol_version__, __jupyter_widgets_base_version__
 
-
 # Based on jupyter_core.paths.envset
 def envset(name, default):
     """Return True if the given environment variable is turned on, otherwise False
@@ -41,12 +31,9 @@ def envset(name, default):
     else:
         return bool(default)
 
-
-
-
 PROTOCOL_VERSION_MAJOR = __protocol_version__.split('.')[0]
 CONTROL_PROTOCOL_VERSION_MAJOR = __control_protocol_version__.split('.')[0]
-JUPYTER_WIDGETS_ECHO = envset('JUPYTER_WIDGETS_ECHO', default=False)
+JUPYTER_WIDGETS_ECHO = envset('JUPYTER_WIDGETS_ECHO', default=True)
 
 def _widget_to_json(x, obj):
     if isinstance(x, dict):
@@ -63,8 +50,8 @@ def _json_to_widget(x, obj):
         return {k: _json_to_widget(v, obj) for k, v in x.items()}
     elif isinstance(x, (list, tuple)):
         return [_json_to_widget(v, obj) for v in x]
-    elif isinstance(x, string_types) and x.startswith('IPY_MODEL_') and x[10:] in Widget.widgets:
-        return Widget.widgets[x[10:]]
+    elif isinstance(x, str) and x.startswith('IPY_MODEL_') and x[10:] in Widget._active_widgets:
+        return Widget._active_widgets[x[10:]]
     else:
         return x
 
@@ -73,10 +60,7 @@ widget_serialization = {
     'to_json': _widget_to_json
 }
 
-if PY3:
-    _binary_types = (memoryview, bytearray, bytes)
-else:
-    _binary_types = (memoryview, bytearray)
+_binary_types = (memoryview, bytearray, bytes)
 
 def _put_buffers(state, buffer_paths, buffers):
     """The inverse of _remove_buffers, except here we modify the existing dict/lists.
@@ -167,18 +151,9 @@ def _buffer_list_equal(a, b):
         # NOTE: Simple ia != ib does not always work as intended, as
         # e.g. memoryview(np.frombuffer(ia, dtype='float32')) !=
         # memoryview(np.frombuffer(b)), since the format info differs.
-        if PY3:
-            # compare without copying
-            if memoryview(ia).cast('B') != memoryview(ib).cast('B'):
-                return False
-        else:
-            # python 2 doesn't have memoryview.cast, so we may have to copy
-            if isinstance(ia, memoryview) and ia.format != 'B':
-                ia = ia.tobytes()
-            if isinstance(ib, memoryview) and ib.format != 'B':
-                ib = ib.tobytes()
-            if ia != ib:
-                return False
+        # Compare without copying.
+        if memoryview(ia).cast('B') != memoryview(ib).cast('B'):
+            return False
     return True
 
 
@@ -231,7 +206,7 @@ class CallbackDispatcher(LoggingHasTraits):
             self.callbacks.append(callback)
 
 def _show_traceback(method):
-    """decorator for showing tracebacks in IPython"""
+    """decorator for showing tracebacks"""
     def m(self, *args, **kwargs):
         try:
             return(method(self, *args, **kwargs))
@@ -244,7 +219,7 @@ def _show_traceback(method):
     return m
 
 
-class WidgetRegistry(object):
+class WidgetRegistry:
 
     def __init__(self):
         self._registry = {}
@@ -284,25 +259,17 @@ class WidgetRegistry(object):
                             for view_name, widget in sorted(vn.items()):
                                     yield (model_module, model_version, model_name, view_module, view_version, view_name), widget
 
-def register(name=''):
-    "For backwards compatibility, we support @register(name) syntax."
-    def reg(widget):
-        """A decorator registering a widget class in the widget registry."""
-        w = widget.class_traits()
-        Widget.widget_types.register(w['_model_module'].default_value,
-                                    w['_model_module_version'].default_value,
-                                    w['_model_name'].default_value,
-                                    w['_view_module'].default_value,
-                                    w['_view_module_version'].default_value,
-                                    w['_view_name'].default_value,
-                                    widget)
-        return widget
-    if isinstance(name, string_types):
-        import warnings
-        warnings.warn("Widget registration using a string name has been deprecated. Widget registration now uses a plain `@register` decorator.", DeprecationWarning)
-        return reg
-    else:
-        return reg(name)
+def register(widget):
+    """A decorator registering a widget class in the widget registry."""
+    w = widget.class_traits()
+    Widget._widget_types.register(w['_model_module'].default_value,
+                                 w['_model_module_version'].default_value,
+                                 w['_model_name'].default_value,
+                                 w['_view_module'].default_value,
+                                 w['_view_module_version'].default_value,
+                                 w['_view_name'].default_value,
+                                 widget)
+    return widget
 
 
 class Widget(LoggingHasTraits):
@@ -312,15 +279,15 @@ class Widget(LoggingHasTraits):
     _widget_construction_callback = None
     _control_comm = None
 
-    # widgets is a dictionary of all active widget objects
-    widgets = {}
+    # _active_widgets is a dictionary of all active widget objects
+    _active_widgets = {}
 
-    # widget_types is a registry of widgets by module, version, and name:
-    widget_types = WidgetRegistry()
+    # _widget_types is a registry of widgets by module, version, and name:
+    _widget_types = WidgetRegistry()
 
     @classmethod
     def close_all(cls):
-        for widget in list(cls.widgets.values()):
+        for widget in list(cls._active_widgets.values()):
             widget.close()
 
     @staticmethod
@@ -362,7 +329,7 @@ class Widget(LoggingHasTraits):
         if method == 'request_states':
             # Send back the full widgets state
             cls.get_manager_state()
-            widgets = cls.widgets.values()
+            widgets = cls._active_widgets.values()
             full_state = {}
             drop_defaults = False
             for widget in widgets:
@@ -392,7 +359,7 @@ class Widget(LoggingHasTraits):
         state = data['state']
 
         # Find the widget class to instantiate in the registered widgets
-        widget_class = Widget.widget_types.get(state['_model_module'],
+        widget_class = Widget._widget_types.get(state['_model_module'],
                                                state['_model_module_version'],
                                                state['_model_name'],
                                                state['_view_module'],
@@ -413,7 +380,7 @@ class Widget(LoggingHasTraits):
         """
         state = {}
         if widgets is None:
-            widgets = Widget.widgets.values()
+            widgets = Widget._active_widgets.values()
         for widget in widgets:
             state[widget.model_id] = widget._get_embed_state(drop_defaults=drop_defaults)
         return {'version_major': 2, 'version_minor': 0, 'state': state}
@@ -465,7 +432,6 @@ class Widget(LoggingHasTraits):
     _property_lock = Dict()
     _holding_sync = False
     _states_to_send = Set()
-    _display_callbacks = Instance(CallbackDispatcher, ())
     _msg_callbacks = Instance(CallbackDispatcher, ())
 
     #-------------------------------------------------------------------------
@@ -474,7 +440,7 @@ class Widget(LoggingHasTraits):
     def __init__(self, **kwargs):
         """Public constructor"""
         self._model_id = kwargs.pop('model_id', None)
-        super(Widget, self).__init__(**kwargs)
+        super().__init__(**kwargs)
 
         Widget._call_widget_constructed(self)
         self.open()
@@ -510,7 +476,7 @@ class Widget(LoggingHasTraits):
         self._model_id = self.model_id
 
         self.comm.on_msg(self._handle_msg)
-        Widget.widgets[self.model_id] = self
+        Widget._active_widgets[self.model_id] = self
 
     @property
     def model_id(self):
@@ -530,10 +496,10 @@ class Widget(LoggingHasTraits):
         When the comm is closed, all of the widget views are automatically
         removed from the front-end."""
         if self.comm is not None:
-            Widget.widgets.pop(self.model_id, None)
+            Widget._active_widgets.pop(self.model_id, None)
             self.comm.close()
             self.comm = None
-            self._ipython_display_ = None
+            self._repr_mimebundle_ = None
 
     def send_state(self, key=None):
         """Sends the widget state, or a piece of it, to the front-end, if it exists.
@@ -570,7 +536,7 @@ class Widget(LoggingHasTraits):
         """
         if key is None:
             keys = self.keys
-        elif isinstance(key, string_types):
+        elif isinstance(key, str):
             keys = [key]
         elif isinstance(key, Iterable):
             keys = key
@@ -581,8 +547,6 @@ class Widget(LoggingHasTraits):
         for k in keys:
             to_json = self.trait_metadata(k, 'to_json', self._trait_to_json)
             value = to_json(getattr(self, k), self)
-            if not PY3 and isinstance(traits[k], Bytes) and isinstance(value, bytes):
-                value = memoryview(value)
             if not drop_defaults or not self._compare(value, traits[k].default_value):
                 state[k] = value
         return state
@@ -602,8 +566,8 @@ class Widget(LoggingHasTraits):
         # Send an echo update message immediately
         if JUPYTER_WIDGETS_ECHO:
             echo_state = {}
-            for attr,value in sync_data.items():
-                if self.trait_metadata(attr, 'echo_update', default=True):
+            for attr, value in sync_data.items():
+                if attr in self.keys and self.trait_metadata(attr, 'echo_update', default=True):
                     echo_state[attr] = value
             if echo_state:
                 echo_state, echo_buffer_paths, echo_buffers = _remove_buffers(echo_state)
@@ -617,7 +581,7 @@ class Widget(LoggingHasTraits):
         # The order of these context managers is important. Properties must
         # be locked when the hold_trait_notification context manager is
         # released and notifications are fired.
-        with self._lock_property(**sync_data), self.hold_trait_notifications():
+        with self.hold_sync(), self._lock_property(**sync_data), self.hold_trait_notifications():
             for name in sync_data:
                 if name in self.keys:
                     from_json = self.trait_metadata(name, 'from_json',
@@ -650,24 +614,9 @@ class Widget(LoggingHasTraits):
             True if the callback should be unregistered."""
         self._msg_callbacks.register_callback(callback, remove=remove)
 
-    def on_displayed(self, callback, remove=False):
-        """(Un)Register a widget displayed callback.
-
-        Parameters
-        ----------
-        callback: method handler
-            Must have a signature of::
-
-                callback(widget, **kwargs)
-
-            kwargs from display are passed through without modification.
-        remove: bool
-            True if the callback should be unregistered."""
-        self._display_callbacks.register_callback(callback, remove=remove)
-
     def add_traits(self, **traits):
         """Dynamically add trait attributes to the Widget."""
-        super(Widget, self).add_traits(**traits)
+        super().add_traits(**traits)
         for name, trait in traits.items():
             if trait.get_metadata('sync'):
                 self.keys.append(name)
@@ -683,7 +632,7 @@ class Widget(LoggingHasTraits):
             if name in self.keys and self._should_send_property(name, getattr(self, name)):
                 # Send new state to front-end
                 self.send_state(key=name)
-        super(Widget, self).notify_change(change)
+        super().notify_change(change)
 
     def __repr__(self):
         return self._gen_repr_from_keys(self._repr_keys())
@@ -734,6 +683,8 @@ class Widget(LoggingHasTraits):
             if (jsonloads(jsondumps(split_value[0])) == split_lock[0]
                 and split_value[1] == split_lock[1]
                 and _buffer_list_equal(split_value[2], split_lock[2])):
+                if self._holding_sync:
+                    self._states_to_send.discard(key)
                 return False
         if self._holding_sync:
             self._states_to_send.add(key)
@@ -772,10 +723,6 @@ class Widget(LoggingHasTraits):
         """Called when a custom msg is received."""
         self._msg_callbacks(self, content, buffers)
 
-    def _handle_displayed(self, **kwargs):
-        """Called when a view has been displayed for this widget instance"""
-        self._display_callbacks(self, **kwargs)
-
     @staticmethod
     def _trait_to_json(x, self):
         """Convert a trait value to json."""
@@ -786,9 +733,7 @@ class Widget(LoggingHasTraits):
         """Convert json values to objects."""
         return x
 
-    def _ipython_display_(self, **kwargs):
-        """Called when `IPython.display.display` is called on the widget."""
-
+    def _repr_mimebundle_(self, **kwargs):
         plaintext = repr(self)
         if len(plaintext) > 110:
             plaintext = plaintext[:110] + '…'
@@ -806,10 +751,7 @@ class Widget(LoggingHasTraits):
                 'version_minor': 0,
                 'model_id': self._model_id
             }
-        display(data, raw=True)
-
-        if self._view_name is not None:
-            self._handle_displayed(**kwargs)
+            return data
 
     def _send(self, msg, buffers=None):
         """Sends a message to the model in the front-end."""
@@ -837,7 +779,7 @@ class Widget(LoggingHasTraits):
     def _gen_repr_from_keys(self, keys):
         class_name = self.__class__.__name__
         signature = ', '.join(
-            '%s=%r' % (key, getattr(self, key))
+            '{}={!r}'.format(key, getattr(self, key))
             for key in keys
         )
-        return '%s(%s)' % (class_name, signature)
+        return '{}({})'.format(class_name, signature)
